import csv
from io import StringIO
from enum import Enum


def load_csv(input_file, col_act, col_pred, col_weight=None, delimiter=",", quotechar="\"", header=True):
    """
    Loads the actual, predicted and weight (if present) values from the CSV file as column lists.

    :param input_file: the CSV file to load actual/predicted labels from
    :type input_file: str
    :param col_act: the 0-based index of the column that contains the actual/ground truth labels
    :type col_act: int
    :param col_pred: the 0-based index of the column that contains the predicted labels
    :type col_pred: int
    :param col_weight: the 0-based index of the (optional) column containing the weights (0-1) for the predictions, default is None (ie a weight 1 is assumed)
    :type col_weight: int
    :param delimiter: the delimiter to use for the CSV file
    :type delimiter: str
    :param quotechar: the quote character to use for the CSV file
    :type quotechar: str
    :param header: whether the CSV file has a header
    :type header: bool
    :return: the actual, predicted and weight (None if not present) columns as lists
    :rtype: tuple
    """
    actual = []
    predicted = []
    weight = None if col_weight is None else []
    with open(input_file, "r") as inputf:
        reader = csv.reader(inputf, delimiter=delimiter, quotechar=quotechar)
        first = True
        for row in reader:
            if first and header:
                first = False
                continue
            actual.append(row[col_act - 1])
            predicted.append(row[col_pred - 1])
            if col_weight is not None:
                weight.append(row[col_weight - 1])
    return actual, predicted, weight


class MatrixType(Enum):
    """
    The types of confusion matrices that can be generated.

    Argparse integration taken from here:
    https://stackoverflow.com/a/55500795/4698227
    """
    COUNTS = 1
    PERCENTAGES = 2
    PERCENTAGES_PER_ROW = 3

    def __str__(self):
        return self.name.lower()

    def __repr__(self):
        return str(self)

    @staticmethod
    def argparse(s):
        try:
            return MatrixType[s.upper()]
        except KeyError:
            return s


class MatrixResult(object):
    """
    Encapsulates the matrix result.
    """

    def __init__(self, data, max_decimals=3):
        """
        Initializes the container.

        :param data: the data, list of lists (row-based)
        :type data: list
        :param max_decimals: the maximum decimals after the decimal point to use in case of float values, -1 for no restrictions
        :type max_decimals: int
        """
        self.data = data
        self.max_decimals = max_decimals
        self.separator = " "

    def _cell_to_string(self, c):
        """
        Turns the cell into a string, takes the maximum number of decimals into account.

        :param c: the cell value to convert
        :type c: object
        :return: the converted value
        :rtype: str
        """
        if (self.max_decimals > -1) and isinstance(c, float):
            cell_format = "%." + str(self.max_decimals) + "f"
            result = str(cell_format % c)
        else:
            result = str(c)
        return result

    def to_raw(self):
        """
        Returns the underlying data as list of list (row-based).

        :return: the underlying data
        :rtype: list
        """
        return self.data

    def to_list(self):
        """
        Returns the matrix result as list of list (row-based).

        :return: the result
        :rtype: list
        """
        result = []
        for r in self.data:
            r_new = []
            for c in r:
                r_new.append(self._cell_to_string(c))
            result.append(r_new)
        return result

    def to_csv(self, output_file=None, delimiter=",", quotechar="\"", quoting=csv.QUOTE_MINIMAL):
        """
        Writes the matrix to a CSV file.

        :param output_file: the CSV file to create
        :type output_file: str
        :param delimiter: the cell delimiter to use (default: ,)
        :type delimiter: str
        :param quotechar: the quoting character to use (default: ")
        :type quotechar: str
        :param quoting: the type of quoting to perform, default is csv.QUOTE_MINIMAL
        :type quoting: int
        :return: None if output file provided, otherwise the generated CSV output
        :rtype: str
        """
        rows = self.to_list()
        if output_file is None:
            outputf = StringIO()
            writer = csv.writer(outputf, delimiter=delimiter, quotechar=quotechar, quoting=quoting)
            for r in rows:
                writer.writerow(r)
            return outputf.getvalue()
        else:
            with open(output_file, "w") as outputf:
                writer = csv.writer(outputf, delimiter=delimiter, quotechar=quotechar, quoting=quoting)
                for r in rows:
                    writer.writerow(r)
            return None

    def to_plaintext(self, output_file=None):
        """
        Turns the matrix result into plain text.

        :param output_file: the CSV file to create
        :type output_file: str
        :return: None if output file provided, otherwise the generated plain text output
        :rtype: str
        """
        widths = []
        num_rows = len(self.data)
        num_cols = 0
        for r in self.data:
            if len(widths) == 0:
                num_cols = len(r)
                for i in range(num_cols):
                    widths.append(0)
            for x in range(num_cols):
                widths[x] = max(widths[x], len(self._cell_to_string(r[x])))

        result = ""
        for y in range(num_rows):
            r = self.data[y]
            for x in range(num_cols):
                value = self._cell_to_string(r[x])
                if (x == 0) or (y == 0):
                    result += value.ljust(widths[x])
                else:
                    result += value.rjust(widths[x])
                if x < num_cols - 1:
                    result += self.separator
            result += "\n"

        if output_file is not None:
            with open(output_file, "w") as outputf:
                outputf.write(result)
            return None
        else:
            return result

    def __str__(self):
        return self.to_plaintext()


class ConfusionMatrix(object):
    """
    For generating a confusion matrix from actual and predicted labels.
    """

    def __init__(self, actual, predicted, weight=None, labels=None,
                 actual_prefix="a: ", predicted_prefix="p: ", corner="x"):
        """
        Initializes the matrix.

        :param actual: the list of actual labels
        :type actual: list
        :param predicted: the list of predicted labels
        :type predicted: list
        :param weight: the list of weights (0-1, float) associated with the predictions
        :type weight: list
        :param labels: the predefined list of labels to use, rather than determining from the actual/predicted labels (sorted)
        :type labels: list
        :param actual_prefix: the prefix to use in the matrix to denote actual cells (left column)
        :type actual_prefix: str
        :param predicted_prefix: the prefix to use in the matrix to denote predicted cells (top row)
        :type predicted_prefix: str
        :param corner: the text to use for the top-left corner
        :type corner: str
        """
        if len(actual) != len(predicted):
            raise Exception("Actual and predicted labels differ in length: %d != %d" % (len(actual), len(predicted)))
        self.actual = actual[:]
        self.actual_prefix = actual_prefix
        self.predicted = predicted[:]
        self.predicted_prefix = predicted_prefix
        self.corner = corner
        self.weight = None
        if weight is not None:
            if len(weight) != len(predicted):
                raise Exception("Weights and predicted labels differ in length: %d != %d" % (len(weight), len(predicted)))
            self.weight = weight[:]
        if labels is not None:
            self.labels = labels[:]
        else:
            all_labels = set()
            all_labels.update(actual)
            all_labels.update(predicted)
            self.labels = list(all_labels)
            self.labels.sort()

    def generate(self, matrix_type=MatrixType.COUNTS, max_decimals=3):
        """
        Generates the confusion matrix and returns it as list of list (row-based).

        :param matrix_type: the type of matrix to generate
        :type matrix_type: MatrixType
        :param max_decimals: the maximum decimals after the decimal point to use for float values, -1 for no restrictions
        :type max_decimals: int
        :return: the generated matrix
        :rtype: MatrixResult
        """
        result = []

        indices = dict()
        for i, l in enumerate(self.labels):
            indices[l] = i + 1

        # header
        row = [self.corner]
        for l in self.labels:
            row.append(self.predicted_prefix + l)
        result.append(row)

        # data
        for l in self.labels:
            row = [self.actual_prefix + l]
            for i in range(len(self.labels)):
                row.append(0)
            result.append(row)

        # fill in counts
        for i in range(len(self.actual)):
            index_act = indices[self.actual[i]]
            index_pred = indices[self.predicted[i]]
            weight = 1 if self.weight is None else self.weight[i]
            result[index_act][index_pred] += weight

        # post-process cells?
        if matrix_type == MatrixType.PERCENTAGES:
            sum = 0.0
            for y in range(len(self.labels)):
                for x in range(len(self.labels)):
                    sum += result[y+1][x+1]
            if sum > 0:
                for y in range(len(self.labels)):
                    for x in range(len(self.labels)):
                        result[y + 1][x + 1] /= sum
        elif matrix_type == MatrixType.PERCENTAGES_PER_ROW:
            for y in range(len(self.labels)):
                sum = 0
                for x in range(len(self.labels)):
                    sum += result[y+1][x+1]
                for x in range(len(self.labels)):
                    result[y + 1][x + 1] /= sum

        return MatrixResult(result, max_decimals=max_decimals)



import argparse
import csv
import traceback
from enum import Enum
from scm.matrix import ConfusionMatrix, MatrixType, load_csv


class OutputFormat(Enum):
    """
    The types of output formats.

    Argparse integration taken from here:
    https://stackoverflow.com/a/55500795/4698227
    """
    CSV = 1
    PLAINTEXT = 2

    def __str__(self):
        return self.name.lower()

    def __repr__(self):
        return str(self)

    @staticmethod
    def argparse(s):
        try:
            return OutputFormat[s.upper()]
        except KeyError:
            return s


def generate(input_file, output_file, col_act, col_pred, col_weight=None, matrix_type=MatrixType.COUNTS,
             delimiter=",", quotechar="\"", header=True, labels=None, prefix_act="a: ", prefix_pred="p: ", corner="x",
             output_format=OutputFormat.CSV, max_decimals=3):
    """
    Generates the confusion matrix from the CSV file. Outputs the result on stdout if no output file provided.

    :param input_file: the CSV file to load actual/predicted labels from
    :type input_file: str
    :param output_file: the (optional) CSV file to write the matrix to
    :type output_file: str
    :param col_act: the 1-based index of the column that contains the actual/ground truth labels
    :type col_act: int
    :param col_pred: the 1-based index of the column that contains the predicted labels
    :type col_pred: int
    :param col_weight: the 1-based index of the (optional) column containing the weight (0-1) for the predictions, default is None
    :type col_weight: int
    :param matrix_type: the type of matrix to generate
    :type matrix_type: MatrixType
    :param delimiter: the delimiter to use for the CSV file
    :type delimiter: str
    :param quotechar: the quote character to use for the CSV file
    :type quotechar: str
    :param header: whether the CSV file has a header
    :type header: bool
    :param labels: the (optional) list of predefined labels to use
    :type labels: list
    :param prefix_act: the prefix to use for the the actual cells (left column)
    :type prefix_act: str
    :param prefix_pred: the prefix to use for the the predicted cells (top row)
    :type prefix_pred: str
    :param corner: the text to print in the top-left corner
    :type corner: str
    :param output_format: the format to use when writing to a file (csv|plaintext)
    :type output_format: OutputFormat
    :param max_decimals: the maximum decimals after the decimal point to use in case of float values, -1 for no restrictions
    :type max_decimals: int
    """

    actual, predicted, weight = load_csv(input_file, col_act, col_pred, col_weight=col_weight,
                                         delimiter=delimiter, quotechar=quotechar, header=header)
    matrix = ConfusionMatrix(actual, predicted, weight=weight, labels=labels,
                             actual_prefix=prefix_act, predicted_prefix=prefix_pred, corner=corner)
    result = matrix.generate(matrix_type=matrix_type, max_decimals=max_decimals)

    if output_file is None:
        print(result.to_plaintext())
    else:
        with open(output_file, "w") as outputf:
            if output_format == OutputFormat.CSV:
                result.to_csv(output_file, delimiter=delimiter, quotechar=quotechar, quoting=csv.QUOTE_MINIMAL)
            elif output_format == OutputFormat.PLAINTEXT:
                outputf.write(result.to_plaintext())
            else:
                raise Exception("Unhandled output format: " + str(output_format))


def main(args=None):
    """
    Performs the matrix generation.
    Use -h to see all options.

    :param args: the command-line arguments to use, uses sys.argv if None
    :type args: list
    """
    parser = argparse.ArgumentParser(
        description='Generates a confusion matrix from a CSV file with actual/predicted label columns.',
        prog="scm-generate",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-i", "--input", dest="input_file", metavar="FILE", required=True, help="the CSV file to load the actual/predicted labels from")
    parser.add_argument("-d", "--delimiter", dest="delimiter", metavar="DELIMITER", required=False, default=",", help="the column delimited in the CSV file")
    parser.add_argument("-q", "--quotechar", dest="quotechar", metavar="CHAR", required=False, default="\"", help="the quote character to use in the CSV file")
    parser.add_argument("-H", "--no_header", action="store_false", dest="header", help="whether the CSV file has no header row")
    parser.add_argument("-o", "--output", dest="output_file", metavar="FILE", required=False, help="the optional CSV file to write the generated matrix to")
    parser.add_argument("-O", "--output_format", dest="output_format", metavar="FORMAT", required=False, default=OutputFormat.CSV, choices=list(OutputFormat), type=OutputFormat.argparse, help="the output format to use when writing to the output file")
    parser.add_argument("-a", "--actual", dest="col_act", metavar="COL", required=False, default=1, type=int, help="the 1-based column index for the actual/ground truth labels")
    parser.add_argument("-A", "--actual_prefix", dest="prefix_act", metavar="PREFIX", required=False, default="a: ", type=str, help="the prefix to use for the labels depicted in the 'actual' labels column")
    parser.add_argument("-p", "--predicted", dest="col_pred", metavar="COL", required=False, default=2, type=int, help="the 1-based column index for the predicted labels")
    parser.add_argument("-P", "--predicted_prefix", dest="prefix_pred", metavar="PREFIX", required=False, default="p: ", type=str, help="the prefix to use for the labels depicted in the 'predicted' labels row")
    parser.add_argument("-w", "--weight", dest="col_weight", metavar="COL", required=False, default=None, type=int, help="the 1-based column index for the weight (0-1) of the predicted label")
    parser.add_argument("-l", "--labels", dest="labels", metavar="LABELS", required=False, default=None, type=str, help="comma-separated list of predefined labels to use (eg if not all labels present in CSV file)")
    parser.add_argument("-C", "--corner", dest="corner", metavar="CORNER", required=False, default="x", type=str, help="the text to print in the top-left corner")
    parser.add_argument("-D", "--max_decimals", dest="max_decimals", metavar="NUM", required=False, default=3, type=int, help="the maximum number of decimals after the decimal point to use in case of float values like percentages")
    parser.add_argument("-t", "--matrix_type", dest="matrix_type", metavar="TYPE", required=False, default=MatrixType.COUNTS, choices=list(MatrixType), type=MatrixType.argparse, help="the type of matrix to generate")
    parsed = parser.parse_args(args=args)
    labels = None if parsed.labels is None else parsed.labels.split(",")
    generate(parsed.input_file, parsed.output_file, parsed.col_act, parsed.col_pred,
             col_weight=parsed.col_weight, matrix_type=parsed.matrix_type,
             delimiter=parsed.delimiter, quotechar=parsed.quotechar, header=parsed.header, labels=labels,
             prefix_act=parsed.prefix_act, prefix_pred=parsed.prefix_pred, corner=parsed.corner,
             output_format=parsed.output_format, max_decimals=parsed.max_decimals)


def sys_main():
    """
    Runs the main function using the system cli arguments, and
    returns a system error code.

    :return: 0 for success, 1 for failure.
    :rtype: int
    """

    try:
        main()
        return 0
    except Exception:
        print(traceback.format_exc())
        return 1


if __name__ == "__main__":
    try:
        main()
    except Exception:
        print(traceback.format_exc())

